@@ -1,4 +1,4 @@
-// Copyright (c) 2018 Microsoft Corporation
+// Copyright (c) 2023 The InterpretML Contributors
 // Licensed under the MIT license.
 // Author: Paul Koch <code@koch.ninja>
 
@@ -7,11 +7,13 @@
 #include <cstring> // memcpy, strcmp
 #include <algorithm> // std::min, std::max
 
-#include "ebm_native.h"
+#include "libebm.h"
 #include "logging.h"
+
+#define ZONE_main
 #include "zones.h"
 
-#include "common_cpp.hpp"
+#include "common.hpp"
 
 #include "ebm_internal.hpp"
 #include "Feature.hpp"
@@ -20,6 +22,7 @@
 #include "BoosterShell.hpp"
 #include "InteractionCore.hpp"
 
+#define R_NO_REMAP
 #include <Rinternals.h>
 #include <R_ext/Visibility.h>
 
@@ -41,20 +44,20 @@ namespace DEFINED_ZONE_NAME {
 
 INLINE_ALWAYS static double ConvertDouble(const SEXP sexp) {
    if(REALSXP != TYPEOF(sexp)) {
-      error("ConvertDouble REALSXP != TYPEOF(sexp)");
+      Rf_error("ConvertDouble REALSXP != TYPEOF(sexp)");
    }
-   if(R_xlen_t { 1 } != xlength(sexp)) {
-      error("ConvertDouble R_xlen_t { 1 } != xlength(sexp)");
+   if(R_xlen_t { 1 } != Rf_xlength(sexp)) {
+      Rf_error("ConvertDouble R_xlen_t { 1 } != Rf_xlength(sexp)");
    }
    return REAL(sexp)[0];
 }
 
 INLINE_ALWAYS static IntEbm ConvertIndex(double index) {
    if(std::isnan(index)) {
-      error("ConvertIndex std::isnan(index)");
+      Rf_error("ConvertIndex std::isnan(index)");
    }
    if(index < 0) {
-      error("ConvertIndex index < 0");
+      Rf_error("ConvertIndex index < 0");
    }
    static constexpr double maxValid = EbmMin(
       double { R_XLEN_T_MAX }, 
@@ -64,7 +67,7 @@ INLINE_ALWAYS static IntEbm ConvertIndex(double index) {
       static_cast<double>(std::numeric_limits<R_xlen_t>::max())
    );
    if(maxValid < index) {
-      error("ConvertIndex maxValid < index");
+      Rf_error("ConvertIndex maxValid < index");
    }
    return static_cast<IntEbm>(index);
 }
@@ -75,7 +78,7 @@ INLINE_ALWAYS static IntEbm ConvertIndex(const SEXP sexp) {
 
 INLINE_ALWAYS static IntEbm ConvertIndexApprox(double index) {
    if(std::isnan(index)) {
-      error("ConvertIndexApprox std::isnan(index)");
+      Rf_error("ConvertIndexApprox std::isnan(index)");
    }
    static constexpr double minValid = EbmMax(
       double { -FLOAT64_TO_INT64_MAX },
@@ -100,20 +103,20 @@ INLINE_ALWAYS static IntEbm ConvertIndexApprox(const SEXP sexp) {
 
 INLINE_ALWAYS static IntEbm ConvertInt(const SEXP sexp) {
    if(INTSXP != TYPEOF(sexp)) {
-      error("ConvertInt INTSXP != TYPEOF(sexp)");
+      Rf_error("ConvertInt INTSXP != TYPEOF(sexp)");
    }
-   if(R_xlen_t { 1 } != xlength(sexp)) {
-      error("ConvertInt R_xlen_t { 1 } != xlength(sexp)");
+   if(R_xlen_t { 1 } != Rf_xlength(sexp)) {
+      Rf_error("ConvertInt R_xlen_t { 1 } != Rf_xlength(sexp)");
    }
    return INTEGER(sexp)[0];
 }
 
 INLINE_ALWAYS static BoolEbm ConvertBool(const SEXP sexp) {
    if(LGLSXP != TYPEOF(sexp)) {
-      error("ConvertBool LGLSXP != TYPEOF(sexp)");
+      Rf_error("ConvertBool LGLSXP != TYPEOF(sexp)");
    }
-   if(R_xlen_t { 1 } != xlength(sexp)) {
-      error("ConvertBool R_xlen_t { 1 } != xlength(sexp)");
+   if(R_xlen_t { 1 } != Rf_xlength(sexp)) {
+      Rf_error("ConvertBool R_xlen_t { 1 } != Rf_xlength(sexp)");
    }
    const Rboolean val = static_cast<Rboolean>(LOGICAL(sexp)[0]);
    if(Rboolean::FALSE == val) {
@@ -122,17 +125,18 @@ INLINE_ALWAYS static BoolEbm ConvertBool(const SEXP sexp) {
    if(Rboolean::TRUE == val) {
       return EBM_TRUE;
    }
-   error("ConvertBool val not a bool");
+   Rf_error("ConvertBool val not a bool");
+   return EBM_FALSE;  // this should never be reached because error uses longjump?!?
 }
 
 static IntEbm CountInts(const SEXP a) {
    EBM_ASSERT(nullptr != a);
    if(INTSXP != TYPEOF(a)) {
-      error("CountInts INTSXP != TYPEOF(a)");
+      Rf_error("CountInts INTSXP != TYPEOF(a)");
    }
-   const R_xlen_t c = xlength(a);
+   const R_xlen_t c = Rf_xlength(a);
    if(IsConvertError<size_t>(c) || IsConvertError<IntEbm>(c)) {
-      error("CountInts IsConvertError<size_t>(c) || IsConvertError<IntEbm>(c)");
+      Rf_error("CountInts IsConvertError<size_t>(c) || IsConvertError<IntEbm>(c)");
    }
    return static_cast<IntEbm>(c);
 }
@@ -140,11 +144,11 @@ static IntEbm CountInts(const SEXP a) {
 static IntEbm CountDoubles(const SEXP a) {
    EBM_ASSERT(nullptr != a);
    if(REALSXP != TYPEOF(a)) {
-      error("CountDoubles REALSXP != TYPEOF(a)");
+      Rf_error("CountDoubles REALSXP != TYPEOF(a)");
    }
-   const R_xlen_t c = xlength(a);
+   const R_xlen_t c = Rf_xlength(a);
    if(IsConvertError<size_t>(c) || IsConvertError<IntEbm>(c)) {
-      error("CountDoubles IsConvertError<size_t>(c) || IsConvertError<IntEbm>(c)");
+      Rf_error("CountDoubles IsConvertError<size_t>(c) || IsConvertError<IntEbm>(c)");
    }
    return static_cast<IntEbm>(c);
 }
@@ -153,7 +157,7 @@ static const IntEbm * ConvertDoublesToIndexes(const IntEbm c, const SEXP a) {
    EBM_ASSERT(0 <= c);
    EBM_ASSERT(nullptr != a);
    if(REALSXP != TYPEOF(a)) {
-      error("ConvertDoublesToIndexes REALSXP != TYPEOF(a)");
+      Rf_error("ConvertDoublesToIndexes REALSXP != TYPEOF(a)");
    }
    IntEbm * aTo = nullptr;
    if(0 < c) {
@@ -182,17 +186,17 @@ static IntEbm CountTotalDimensions(const size_t cTerms, const IntEbm * const acT
       do {
          const IntEbm countDimensions = *pcTermDimensions;
          if(IsConvertError<size_t>(countDimensions)) {
-            error("CountTotalDimensions IsConvertError<size_t>(countDimensions)");
+            Rf_error("CountTotalDimensions IsConvertError<size_t>(countDimensions)");
          }
          const size_t cDimensions = static_cast<size_t>(countDimensions);
          if(IsAddError(cTotalDimensions, cDimensions)) {
-            error("CountTotalDimensions IsAddError(cTotalDimensions, cDimensions)");
+            Rf_error("CountTotalDimensions IsAddError(cTotalDimensions, cDimensions)");
          }
          cTotalDimensions += cDimensions;
          ++pcTermDimensions;
       } while(pcTermDimensionsEnd != pcTermDimensions);
       if(IsConvertError<IntEbm>(cTotalDimensions)) {
-         error("CountTotalDimensions IsConvertError<IntEbm>(cTotalDimensions)");
+         Rf_error("CountTotalDimensions IsConvertError<IntEbm>(cTotalDimensions)");
       }
    }
    return static_cast<IntEbm>(cTotalDimensions);
@@ -289,10 +293,10 @@ SEXP CutQuantile_R(SEXP featureVals, SEXP minSamplesBin, SEXP isRounded, SEXP co
       aCutsLowerBoundInclusive
    );
    if(Error_None != err) {
-      error("CutQuantile returned error code: %" ErrorEbmPrintf, err);
+      Rf_error("CutQuantile returned error code: %" ErrorEbmPrintf, err);
    }
 
-   const SEXP ret = PROTECT(allocVector(REALSXP, static_cast<R_xlen_t>(cCuts)));
+   const SEXP ret = PROTECT(Rf_allocVector(REALSXP, static_cast<R_xlen_t>(cCuts)));
 
    // we've allocated this memory, so it should be reachable, so these numbers should multiply
    EBM_ASSERT(!IsMultiplyError(sizeof(*aCutsLowerBoundInclusive), static_cast<size_t>(cCuts)));
@@ -302,7 +306,7 @@ SEXP CutQuantile_R(SEXP featureVals, SEXP minSamplesBin, SEXP isRounded, SEXP co
       const double * pCutsLowerBoundInclusive = aCutsLowerBoundInclusive;
       const double * const pCutsLowerBoundInclusiveEnd = aCutsLowerBoundInclusive + static_cast<size_t>(cCuts);
       do {
-         *pRet = static_cast<double>(*pCutsLowerBoundInclusive);
+         *pRet = *pCutsLowerBoundInclusive;
          ++pRet;
          ++pCutsLowerBoundInclusive;
       } while(pCutsLowerBoundInclusiveEnd != pCutsLowerBoundInclusive);
@@ -324,13 +328,13 @@ SEXP Discretize_R(SEXP featureVals, SEXP cutsLowerBoundInclusive, SEXP binIndexe
    if(SAFE_FLOAT64_AS_INT64_MAX - 2 < cCuts) {
       // if the number of cuts is low enough, we don't need to check if the bin indexes below exceed our safe float max
       // the highest bin index is +2 from the number of cuts, although the # of bins is 1 higher but we're ok with that
-      error("Discretize_R SAFE_FLOAT64_AS_INT64_MAX - 2 < cCuts");
+      Rf_error("Discretize_R SAFE_FLOAT64_AS_INT64_MAX - 2 < cCuts");
    }
    const double * const aCutsLowerBoundInclusive = REAL(cutsLowerBoundInclusive);
 
    const IntEbm cBinIndexesOut = CountDoubles(binIndexesOut);
    if(cSamples != cBinIndexesOut) {
-      error("Discretize_R cSamples != cBinIndexesOut");
+      Rf_error("Discretize_R cSamples != cBinIndexesOut");
    }
 
    if(0 != cSamples) {
@@ -340,7 +344,7 @@ SEXP Discretize_R(SEXP featureVals, SEXP cutsLowerBoundInclusive, SEXP binIndexe
 
       const ErrorEbm err = Discretize(cSamples, aFeatureVals, cCuts, aCutsLowerBoundInclusive, aiBins);
       if(Error_None != err) {
-         error("Discretize returned error code: %" ErrorEbmPrintf, err);
+         Rf_error("Discretize returned error code: %" ErrorEbmPrintf, err);
       }
 
       double * pBinIndexesOut = REAL(binIndexesOut);
@@ -368,13 +372,13 @@ SEXP MeasureDataSetHeader_R(SEXP countFeatures, SEXP countWeights, SEXP countTar
 
    const IntEbm countBytes = MeasureDataSetHeader(cFeatures, cWeights, cTargets);
    if(countBytes < 0) {
-      error("MeasureDataSetHeader_R MeasureDataSetHeader returned error code: %" ErrorEbmPrintf, static_cast<ErrorEbm>(countBytes));
+      Rf_error("MeasureDataSetHeader_R MeasureDataSetHeader returned error code: %" ErrorEbmPrintf, static_cast<ErrorEbm>(countBytes));
    }
    if(SAFE_FLOAT64_AS_INT64_MAX < countBytes) {
-      error("MeasureDataSetHeader_R SAFE_FLOAT64_AS_INT64_MAX < countBytes");
+      Rf_error("MeasureDataSetHeader_R SAFE_FLOAT64_AS_INT64_MAX < countBytes");
    }
 
-   const SEXP ret = PROTECT(allocVector(REALSXP, R_xlen_t { 1 }));
+   const SEXP ret = PROTECT(Rf_allocVector(REALSXP, R_xlen_t { 1 }));
    REAL(ret)[0] = static_cast<double>(countBytes);
    UNPROTECT(1);
    return ret;
@@ -404,13 +408,13 @@ SEXP MeasureFeature_R(SEXP countBins, SEXP isMissing, SEXP isUnknown, SEXP isNom
       aiBins
    );
    if(countBytes < 0) {
-      error("MeasureFeature_R MeasureFeature returned error code: %" ErrorEbmPrintf, static_cast<ErrorEbm>(countBytes));
+      Rf_error("MeasureFeature_R MeasureFeature returned error code: %" ErrorEbmPrintf, static_cast<ErrorEbm>(countBytes));
    }
    if(SAFE_FLOAT64_AS_INT64_MAX < countBytes) {
-      error("MeasureFeature_R SAFE_FLOAT64_AS_INT64_MAX < countBytes");
+      Rf_error("MeasureFeature_R SAFE_FLOAT64_AS_INT64_MAX < countBytes");
    }
 
-   const SEXP ret = PROTECT(allocVector(REALSXP, R_xlen_t { 1 }));
+   const SEXP ret = PROTECT(Rf_allocVector(REALSXP, R_xlen_t { 1 }));
    REAL(ret)[0] = static_cast<double>(countBytes);
    UNPROTECT(1);
    return ret;
@@ -431,13 +435,13 @@ SEXP MeasureClassificationTarget_R(SEXP countClasses, SEXP targets) {
       aTargets
    );
    if(countBytes < 0) {
-      error("MeasureClassificationTarget_R MeasureClassificationTarget returned error code: %" ErrorEbmPrintf, static_cast<ErrorEbm>(countBytes));
+      Rf_error("MeasureClassificationTarget_R MeasureClassificationTarget returned error code: %" ErrorEbmPrintf, static_cast<ErrorEbm>(countBytes));
    }
    if(SAFE_FLOAT64_AS_INT64_MAX < countBytes) {
-      error("MeasureClassificationTarget_R SAFE_FLOAT64_AS_INT64_MAX < countBytes");
+      Rf_error("MeasureClassificationTarget_R SAFE_FLOAT64_AS_INT64_MAX < countBytes");
    }
 
-   const SEXP ret = PROTECT(allocVector(REALSXP, R_xlen_t { 1 }));
+   const SEXP ret = PROTECT(Rf_allocVector(REALSXP, R_xlen_t { 1 }));
    REAL(ret)[0] = static_cast<double>(countBytes);
    UNPROTECT(1);
    return ret;
@@ -485,7 +489,7 @@ SEXP FillDataSetHeader_R(
    const IntEbm cBytesAllocated = ConvertIndex(countBytesAllocated);
 
    if(EXTPTRSXP != TYPEOF(fillMemWrapped)) {
-      error("FillDataSetHeader_R EXTPTRSXP != TYPEOF(fillMemWrapped)");
+      Rf_error("FillDataSetHeader_R EXTPTRSXP != TYPEOF(fillMemWrapped)");
    }
    void * const pDataset = R_ExternalPtrAddr(fillMemWrapped);
 
@@ -497,7 +501,7 @@ SEXP FillDataSetHeader_R(
       pDataset
    );
    if(Error_None != err) {
-      error("FillDataSetHeader returned error code: %" ErrorEbmPrintf, err);
+      Rf_error("FillDataSetHeader returned error code: %" ErrorEbmPrintf, err);
    }
 
    return R_NilValue;
@@ -531,7 +535,7 @@ SEXP FillFeature_R(
    const IntEbm cBytesAllocated = ConvertIndex(countBytesAllocated);
 
    if(EXTPTRSXP != TYPEOF(fillMemWrapped)) {
-      error("FillFeature_R EXTPTRSXP != TYPEOF(fillMemWrapped)");
+      Rf_error("FillFeature_R EXTPTRSXP != TYPEOF(fillMemWrapped)");
    }
    void * const pDataset = R_ExternalPtrAddr(fillMemWrapped);
 
@@ -546,7 +550,7 @@ SEXP FillFeature_R(
       pDataset
    );
    if(Error_None != err) {
-      error("FillFeature returned error code: %" ErrorEbmPrintf, err);
+      Rf_error("FillFeature returned error code: %" ErrorEbmPrintf, err);
    }
 
    return R_NilValue;
@@ -566,7 +570,7 @@ SEXP FillClassificationTarget_R(SEXP countClasses, SEXP targets, SEXP countBytes
    const IntEbm cBytesAllocated = ConvertIndex(countBytesAllocated);
 
    if(EXTPTRSXP != TYPEOF(fillMemWrapped)) {
-      error("FillClassificationTarget_R EXTPTRSXP != TYPEOF(fillMemWrapped)");
+      Rf_error("FillClassificationTarget_R EXTPTRSXP != TYPEOF(fillMemWrapped)");
    }
    void * const pDataset = R_ExternalPtrAddr(fillMemWrapped);
 
@@ -578,7 +582,7 @@ SEXP FillClassificationTarget_R(SEXP countClasses, SEXP targets, SEXP countBytes
       pDataset
    );
    if(Error_None != err) {
-      error("FillClassificationTarget returned error code: %" ErrorEbmPrintf, err);
+      Rf_error("FillClassificationTarget returned error code: %" ErrorEbmPrintf, err);
    }
 
    return R_NilValue;
@@ -593,7 +597,7 @@ SEXP SampleWithoutReplacement_R(SEXP rng, SEXP countTrainingSamples, SEXP countV
    void * pRng = nullptr;
    if(NILSXP != TYPEOF(rng)) {
       if(EXTPTRSXP != TYPEOF(rng)) {
-         error("SampleWithoutReplacement_R EXTPTRSXP != TYPEOF(rng)");
+         Rf_error("SampleWithoutReplacement_R EXTPTRSXP != TYPEOF(rng)");
       }
       pRng = R_ExternalPtrAddr(rng);
    }
@@ -601,13 +605,13 @@ SEXP SampleWithoutReplacement_R(SEXP rng, SEXP countTrainingSamples, SEXP countV
    const IntEbm cTrainingSamples = ConvertIndex(countTrainingSamples);
    const IntEbm cValidationSamples = ConvertIndex(countValidationSamples);
    if(IsAddError(static_cast<size_t>(cTrainingSamples), static_cast<size_t>(cValidationSamples))) {
-      error("SampleWithoutReplacement_R IsAddError(static_cast<size_t>(cTrainingSamples), static_cast<size_t>(cValidationSamples))");
+      Rf_error("SampleWithoutReplacement_R IsAddError(static_cast<size_t>(cTrainingSamples), static_cast<size_t>(cValidationSamples))");
    }
 
    const size_t cSamples = static_cast<size_t>(CountInts(bagOut));
 
    if(static_cast<size_t>(cTrainingSamples) + static_cast<size_t>(cValidationSamples) != cSamples) {
-      error("SampleWithoutReplacement_R static_cast<size_t>(cTrainingSamples) + static_cast<size_t>(cValidationSamples) != cSamples");
+      Rf_error("SampleWithoutReplacement_R static_cast<size_t>(cTrainingSamples) + static_cast<size_t>(cValidationSamples) != cSamples");
    }
 
    if(0 != cSamples) {
@@ -622,7 +626,7 @@ SEXP SampleWithoutReplacement_R(SEXP rng, SEXP countTrainingSamples, SEXP countV
          aBag
       );
       if(Error_None != err) {
-         error("SampleWithoutReplacementFillDataSetHeader returned error code: %" ErrorEbmPrintf, err);
+         Rf_error("SampleWithoutReplacementFillDataSetHeader returned error code: %" ErrorEbmPrintf, err);
       }
 
       int32_t * pSampleReplicationOut = INTEGER(bagOut);
@@ -631,7 +635,7 @@ SEXP SampleWithoutReplacement_R(SEXP rng, SEXP countTrainingSamples, SEXP countV
       do {
          const BagEbm replication = *pSampleReplication;
          if(IsConvertError<int32_t>(replication)) {
-            error("SampleWithoutReplacement_R IsConvertError<int32_t>(replication)");
+            Rf_error("SampleWithoutReplacement_R IsConvertError<int32_t>(replication)");
          }
          *pSampleReplicationOut = static_cast<int32_t>(replication);
          ++pSampleReplicationOut;
@@ -663,13 +667,13 @@ SEXP CreateBooster_R(
    void * pRng = nullptr;
    if(NILSXP != TYPEOF(rng)) {
       if(EXTPTRSXP != TYPEOF(rng)) {
-         error("CreateBooster_R EXTPTRSXP != TYPEOF(rng)");
+         Rf_error("CreateBooster_R EXTPTRSXP != TYPEOF(rng)");
       }
       pRng = R_ExternalPtrAddr(rng);
    }
 
    if(EXTPTRSXP != TYPEOF(dataSetWrapped)) {
-      error("CreateBooster_R EXTPTRSXP != TYPEOF(dataSetWrapped)");
+      Rf_error("CreateBooster_R EXTPTRSXP != TYPEOF(dataSetWrapped)");
    }
    const void * pDataSet = R_ExternalPtrAddr(dataSetWrapped);
 
@@ -680,7 +684,7 @@ SEXP CreateBooster_R(
 
    err = ExtractDataSetHeader(pDataSet, &countSamples, &unused1, &unused2, &unused3);
    if(Error_None != err) {
-      error("ExtractDataSetHeader returned error code: %" ErrorEbmPrintf, err);
+      Rf_error("ExtractDataSetHeader returned error code: %" ErrorEbmPrintf, err);
    }
    const size_t cSamples = static_cast<size_t>(countSamples); // we trust our internal code that this is convertible
 
@@ -689,7 +693,7 @@ SEXP CreateBooster_R(
    if(NILSXP != TYPEOF(bag)) {
       const size_t cSamplesVerify = static_cast<size_t>(CountInts(bag));
       if(cSamples != cSamplesVerify) {
-         error("CreateBooster_R cSamples != cSamplesVerify");
+         Rf_error("CreateBooster_R cSamples != cSamplesVerify");
       }
 
       aBag = reinterpret_cast<BagEbm *>(R_alloc(cSamples, static_cast<int>(sizeof(BagEbm))));
@@ -703,7 +707,7 @@ SEXP CreateBooster_R(
       do {
          const int32_t replication = *pSampleReplicationR;
          if(IsConvertError<BagEbm>(replication)) {
-            error("CreateBooster_R IsConvertError<BagEbm>(replication)");
+            Rf_error("CreateBooster_R IsConvertError<BagEbm>(replication)");
          }
          if(0 != replication) {
             ++cExpectedInitScores;
@@ -719,7 +723,7 @@ SEXP CreateBooster_R(
       const IntEbm countInitScores = CountDoubles(initScores);
       size_t cInitScores = static_cast<size_t>(countInitScores);
       if(cInitScores != cExpectedInitScores) {
-         error("CreateBooster_R cInitScores != cExpectedInitScores");
+         Rf_error("CreateBooster_R cInitScores != cExpectedInitScores");
       }
       aInitScores = REAL(initScores);
    }
@@ -730,7 +734,7 @@ SEXP CreateBooster_R(
 
    const IntEbm cTotalDimensionsActual = CountDoubles(featureIndexes);
    if(cTotalDimensionsActual != cTotalDimensionsCheck) {
-      error("CreateBooster_R cTotalDimensionsActual != cTotalDimensionsCheck");
+      Rf_error("CreateBooster_R cTotalDimensionsActual != cTotalDimensionsCheck");
    }
    const IntEbm * const aiTermFeatures = ConvertDoublesToIndexes(cTotalDimensionsActual, featureIndexes);
 
@@ -746,11 +750,14 @@ SEXP CreateBooster_R(
       acTermDimensions,
       aiTermFeatures,
       cInnerBags,
+      CreateBoosterFlags_Default,
+      AccelerationFlags_ALL,
+      "log_loss",
       nullptr,
       &boosterHandle
    );
    if(Error_None != err || nullptr == boosterHandle) {
-      error("CreateBooster returned error code: %" ErrorEbmPrintf, err);
+      Rf_error("CreateBooster returned error code: %" ErrorEbmPrintf, err);
    }
 
    SEXP boosterHandleWrapped = R_MakeExternalPtr(static_cast<void *>(boosterHandle), R_NilValue, R_NilValue); // makes an EXTPTRSXP
@@ -772,46 +779,46 @@ SEXP GenerateTermUpdate_R(
    SEXP boosterHandleWrapped,
    SEXP indexTerm,
    SEXP learningRate,
-   SEXP minSamplesLeaf,
+   SEXP minHessian,
    SEXP leavesMax
 ) {
    EBM_ASSERT(nullptr != rng);
    EBM_ASSERT(nullptr != boosterHandleWrapped);
    EBM_ASSERT(nullptr != indexTerm);
    EBM_ASSERT(nullptr != learningRate);
-   EBM_ASSERT(nullptr != minSamplesLeaf);
+   EBM_ASSERT(nullptr != minHessian);
    EBM_ASSERT(nullptr != leavesMax);
 
    void * pRng = nullptr;
    if(NILSXP != TYPEOF(rng)) {
       if(EXTPTRSXP != TYPEOF(rng)) {
-         error("GenerateTermUpdate_R EXTPTRSXP != TYPEOF(rng)");
+         Rf_error("GenerateTermUpdate_R EXTPTRSXP != TYPEOF(rng)");
       }
       pRng = R_ExternalPtrAddr(rng);
    }
 
    if(EXTPTRSXP != TYPEOF(boosterHandleWrapped)) {
-      error("GenerateTermUpdate_R EXTPTRSXP != TYPEOF(boosterHandleWrapped)");
+      Rf_error("GenerateTermUpdate_R EXTPTRSXP != TYPEOF(boosterHandleWrapped)");
    }
    const BoosterHandle boosterHandle = static_cast<BoosterHandle>(R_ExternalPtrAddr(boosterHandleWrapped));
    BoosterShell * const pBoosterShell = BoosterShell::GetBoosterShellFromHandle(boosterHandle);
    if(nullptr == pBoosterShell) {
-      error("GenerateTermUpdate_R nullptr == pBoosterShell");
+      Rf_error("GenerateTermUpdate_R nullptr == pBoosterShell");
    }
 
    const IntEbm iTerm = ConvertIndex(indexTerm);
 
    const double learningRateLocal = ConvertDouble(learningRate);
 
-   const IntEbm samplesLeafMin = ConvertIndexApprox(minSamplesLeaf);
+   const double hessianMin = ConvertDouble(minHessian);
 
    const IntEbm cDimensions = CountDoubles(leavesMax);
    const IntEbm * const aLeavesMax = ConvertDoublesToIndexes(cDimensions, leavesMax);
    if(pBoosterShell->GetBoosterCore()->GetCountTerms() <= static_cast<size_t>(iTerm)) {
-      error("GenerateTermUpdate_R pBoosterShell->GetBoosterCore()->GetCountTerms() <= static_cast<size_t>(iTerm)");
+      Rf_error("GenerateTermUpdate_R pBoosterShell->GetBoosterCore()->GetCountTerms() <= static_cast<size_t>(iTerm)");
    }
    if(static_cast<size_t>(cDimensions) < pBoosterShell->GetBoosterCore()->GetTerms()[static_cast<size_t>(iTerm)]->GetCountDimensions()) {
-      error("GenerateTermUpdate_R static_cast<size_t>(cDimensions) < pBoosterShell->GetBoosterCore()->GetTerms()[static_cast<size_t>(iTerm)]->GetCountDimensions()");
+      Rf_error("GenerateTermUpdate_R static_cast<size_t>(cDimensions) < pBoosterShell->GetBoosterCore()->GetTerms()[static_cast<size_t>(iTerm)]->GetCountDimensions()");
    }
 
    double avgGain;
@@ -820,18 +827,23 @@ SEXP GenerateTermUpdate_R(
       pRng,
       boosterHandle,
       iTerm,
-      BoostFlags_Default,
+      TermBoostFlags_Default,
       learningRateLocal,
-      samplesLeafMin,
+      0,
+      hessianMin,
+      0,
+      0,
+      0,
       aLeavesMax,
+      nullptr,
       &avgGain
    );
    if(Error_None != err) {
-      error("GenerateTermUpdate returned error code: %" ErrorEbmPrintf, err);
+      Rf_error("GenerateTermUpdate returned error code: %" ErrorEbmPrintf, err);
    }
 
-   SEXP ret = PROTECT(allocVector(REALSXP, R_xlen_t { 1 }));
-   REAL(ret)[0] = static_cast<double>(avgGain);
+   SEXP ret = PROTECT(Rf_allocVector(REALSXP, R_xlen_t { 1 }));
+   REAL(ret)[0] = avgGain;
    UNPROTECT(1);
    return ret;
 }
@@ -840,7 +852,7 @@ SEXP ApplyTermUpdate_R(SEXP boosterHandleWrapped) {
    EBM_ASSERT(nullptr != boosterHandleWrapped);
 
    if(EXTPTRSXP != TYPEOF(boosterHandleWrapped)) {
-      error("ApplyTermUpdate_R EXTPTRSXP != TYPEOF(boosterHandleWrapped)");
+      Rf_error("ApplyTermUpdate_R EXTPTRSXP != TYPEOF(boosterHandleWrapped)");
    }
    const BoosterHandle boosterHandle = static_cast<BoosterHandle>(R_ExternalPtrAddr(boosterHandleWrapped));
    // we don't use boosterHandle in this function, so let ApplyTermUpdate check if it's null or invalid
@@ -848,11 +860,11 @@ SEXP ApplyTermUpdate_R(SEXP boosterHandleWrapped) {
    double avgValidationMetric;
    const ErrorEbm err = ApplyTermUpdate(boosterHandle, &avgValidationMetric);
    if(Error_None != err) {
-      error("ApplyTermUpdate returned error code: %" ErrorEbmPrintf, err);
+      Rf_error("ApplyTermUpdate returned error code: %" ErrorEbmPrintf, err);
    }
 
-   SEXP ret = PROTECT(allocVector(REALSXP, R_xlen_t { 1 }));
-   REAL(ret)[0] = static_cast<double>(avgValidationMetric);
+   SEXP ret = PROTECT(Rf_allocVector(REALSXP, R_xlen_t { 1 }));
+   REAL(ret)[0] = avgValidationMetric;
    UNPROTECT(1);
    return ret;
 }
@@ -862,39 +874,41 @@ SEXP GetBestTermScores_R(SEXP boosterHandleWrapped, SEXP indexTerm) {
    EBM_ASSERT(nullptr != indexTerm); // shouldn't be possible
 
    if(EXTPTRSXP != TYPEOF(boosterHandleWrapped)) {
-      error("GetBestTermScores_R EXTPTRSXP != TYPEOF(boosterHandleWrapped)");
+      Rf_error("GetBestTermScores_R EXTPTRSXP != TYPEOF(boosterHandleWrapped)");
    }
    const BoosterHandle boosterHandle = static_cast<BoosterHandle>(R_ExternalPtrAddr(boosterHandleWrapped));
    BoosterShell * const pBoosterShell = BoosterShell::GetBoosterShellFromHandle(boosterHandle);
    if(nullptr == pBoosterShell) {
-      error("GetBestTermScores_R nullptr == pBoosterShell");
+      Rf_error("GetBestTermScores_R nullptr == pBoosterShell");
    }
    BoosterCore * const pBoosterCore = pBoosterShell->GetBoosterCore();
 
    const IntEbm iTerm = ConvertIndex(indexTerm);
 
    if(pBoosterCore->GetCountTerms() <= static_cast<size_t>(iTerm)) {
-      error("GetBestTermScores_R pBoosterCore->GetCountTerms() <= static_cast<size_t>(iTerm)");
+      Rf_error("GetBestTermScores_R pBoosterCore->GetCountTerms() <= static_cast<size_t>(iTerm)");
    }
 
-   size_t cTensorScores = GetCountScores(pBoosterCore->GetCountClasses());
-   const Term * const pTerm = pBoosterCore->GetTerms()[static_cast<size_t>(iTerm)];
-   const size_t cDimensions = pTerm->GetCountDimensions();
-   if(0 != cDimensions) {
-      const FeatureBoosting * const * ppFeature = pTerm->GetFeatures();
-      const FeatureBoosting * const * const ppFeaturesEnd = &ppFeature[cDimensions];
-      do {
-         const FeatureBoosting * const pFeature = *ppFeature;
-         const size_t cBins = pFeature->GetCountBins();
-         EBM_ASSERT(!IsMultiplyError(cTensorScores, cBins)); // we've allocated this memory, so it should be reachable, so these numbers should multiply
-         cTensorScores *= cBins;
-         ++ppFeature;
-      } while(ppFeaturesEnd != ppFeature);
+   size_t cTensorScores = pBoosterCore->GetCountScores();
+   if(size_t { 0 } != cTensorScores) {
+      const Term * const pTerm = pBoosterCore->GetTerms()[static_cast<size_t>(iTerm)];
+      const size_t cDimensions = pTerm->GetCountDimensions();
+      if(0 != cDimensions) {
+         const TermFeature * pTermFeature = pTerm->GetTermFeatures();
+         const TermFeature * const pTermFeaturesEnd = &pTermFeature[cDimensions];
+         do {
+            const FeatureBoosting * const pFeature = pTermFeature->m_pFeature;
+            const size_t cBins = pFeature->GetCountBins();
+            EBM_ASSERT(!IsMultiplyError(cTensorScores, cBins)); // we've allocated this memory, so it should be reachable, so these numbers should multiply
+            cTensorScores *= cBins;
+            ++pTermFeature;
+         } while(pTermFeaturesEnd != pTermFeature);
+      }
+      if(IsConvertError<R_xlen_t>(cTensorScores)) {
+         Rf_error("GetBestTermScores_R IsConvertError<R_xlen_t>(cTensorScores)");
+      }
    }
-   if(IsConvertError<R_xlen_t>(cTensorScores)) {
-      error("GetBestTermScores_R IsConvertError<R_xlen_t>(cTensorScores)");
-   }
-   SEXP ret = PROTECT(allocVector(REALSXP, static_cast<R_xlen_t>(cTensorScores)));
+   SEXP ret = PROTECT(Rf_allocVector(REALSXP, static_cast<R_xlen_t>(cTensorScores)));
    EBM_ASSERT(!IsMultiplyError(sizeof(double), cTensorScores)); // we've allocated this memory, so it should be reachable, so these numbers should multiply
 
    const ErrorEbm err = GetBestTermScores(boosterHandle, iTerm, REAL(ret));
@@ -902,7 +916,7 @@ SEXP GetBestTermScores_R(SEXP boosterHandleWrapped, SEXP indexTerm) {
    UNPROTECT(1);
 
    if(Error_None != err) {
-      error("GetBestTermScores returned error code: %" ErrorEbmPrintf, err);
+      Rf_error("GetBestTermScores returned error code: %" ErrorEbmPrintf, err);
    }
    return ret;
 }
@@ -912,39 +926,41 @@ SEXP GetCurrentTermScores_R(SEXP boosterHandleWrapped, SEXP indexTerm) {
    EBM_ASSERT(nullptr != indexTerm); // shouldn't be possible
 
    if(EXTPTRSXP != TYPEOF(boosterHandleWrapped)) {
-      error("GetCurrentTermScores_R EXTPTRSXP != TYPEOF(boosterHandleWrapped)");
+      Rf_error("GetCurrentTermScores_R EXTPTRSXP != TYPEOF(boosterHandleWrapped)");
    }
    const BoosterHandle boosterHandle = static_cast<BoosterHandle>(R_ExternalPtrAddr(boosterHandleWrapped));
    BoosterShell * const pBoosterShell = BoosterShell::GetBoosterShellFromHandle(boosterHandle);
    if(nullptr == pBoosterShell) {
-      error("GetCurrentTermScores_R nullptr == pBoosterShell");
+      Rf_error("GetCurrentTermScores_R nullptr == pBoosterShell");
    }
    BoosterCore * const pBoosterCore = pBoosterShell->GetBoosterCore();
 
    const IntEbm iTerm = ConvertIndex(indexTerm);
 
    if(pBoosterCore->GetCountTerms() <= static_cast<size_t>(iTerm)) {
-      error("GetCurrentTermScores_R pBoosterCore->GetCountTerms() <= static_cast<size_t>(iTerm)");
+      Rf_error("GetCurrentTermScores_R pBoosterCore->GetCountTerms() <= static_cast<size_t>(iTerm)");
    }
 
-   size_t cTensorScores = GetCountScores(pBoosterCore->GetCountClasses());
-   const Term * const pTerm = pBoosterCore->GetTerms()[static_cast<size_t>(iTerm)];
-   const size_t cDimensions = pTerm->GetCountDimensions();
-   if(0 != cDimensions) {
-      const FeatureBoosting * const * ppFeature = pTerm->GetFeatures();
-      const FeatureBoosting * const * const ppFeaturesEnd = &ppFeature[cDimensions];
-      do {
-         const FeatureBoosting * const pFeature = *ppFeature;
-         const size_t cBins = pFeature->GetCountBins();
-         EBM_ASSERT(!IsMultiplyError(cTensorScores, cBins)); // we've allocated this memory, so it should be reachable, so these numbers should multiply
-         cTensorScores *= cBins;
-         ++ppFeature;
-      } while(ppFeaturesEnd != ppFeature);
+   size_t cTensorScores = pBoosterCore->GetCountScores();
+   if(ptrdiff_t { 0 } != cTensorScores) {
+      const Term * const pTerm = pBoosterCore->GetTerms()[static_cast<size_t>(iTerm)];
+      const size_t cDimensions = pTerm->GetCountDimensions();
+      if(0 != cDimensions) {
+         const TermFeature * pTermFeature = pTerm->GetTermFeatures();
+         const TermFeature * const pTermFeaturesEnd = &pTermFeature[cDimensions];
+         do {
+            const FeatureBoosting * const pFeature = pTermFeature->m_pFeature;
+            const size_t cBins = pFeature->GetCountBins();
+            EBM_ASSERT(!IsMultiplyError(cTensorScores, cBins)); // we've allocated this memory, so it should be reachable, so these numbers should multiply
+            cTensorScores *= cBins;
+            ++pTermFeature;
+         } while(pTermFeaturesEnd != pTermFeature);
+      }
+      if(IsConvertError<R_xlen_t>(cTensorScores)) {
+         Rf_error("GetCurrentTermScores_R IsConvertError<R_xlen_t>(cTensorScores)");
+      }
    }
-   if(IsConvertError<R_xlen_t>(cTensorScores)) {
-      error("GetCurrentTermScores_R IsConvertError<R_xlen_t>(cTensorScores)");
-   }
-   SEXP ret = PROTECT(allocVector(REALSXP, static_cast<R_xlen_t>(cTensorScores)));
+   SEXP ret = PROTECT(Rf_allocVector(REALSXP, static_cast<R_xlen_t>(cTensorScores)));
    EBM_ASSERT(!IsMultiplyError(sizeof(double), cTensorScores)); // we've allocated this memory, so it should be reachable, so these numbers should multiply
 
    const ErrorEbm err = GetCurrentTermScores(boosterHandle, iTerm, REAL(ret));
@@ -952,7 +968,7 @@ SEXP GetCurrentTermScores_R(SEXP boosterHandleWrapped, SEXP indexTerm) {
    UNPROTECT(1);
 
    if(Error_None != err) {
-      error("GetCurrentTermScores returned error code: %" ErrorEbmPrintf, err);
+      Rf_error("GetCurrentTermScores returned error code: %" ErrorEbmPrintf, err);
    }
    return ret;
 }
@@ -965,7 +981,7 @@ SEXP CreateInteractionDetector_R(SEXP dataSetWrapped, SEXP bag, SEXP initScores)
    ErrorEbm err;
 
    if(EXTPTRSXP != TYPEOF(dataSetWrapped)) {
-      error("CreateInteractionDetector_R EXTPTRSXP != TYPEOF(dataSetWrapped)");
+      Rf_error("CreateInteractionDetector_R EXTPTRSXP != TYPEOF(dataSetWrapped)");
    }
    const void * pDataSet = R_ExternalPtrAddr(dataSetWrapped);
 
@@ -976,7 +992,7 @@ SEXP CreateInteractionDetector_R(SEXP dataSetWrapped, SEXP bag, SEXP initScores)
 
    err = ExtractDataSetHeader(pDataSet, &countSamples, &unused1, &unused2, &unused3);
    if(Error_None != err) {
-      error("ExtractDataSetHeader returned error code: %" ErrorEbmPrintf, err);
+      Rf_error("ExtractDataSetHeader returned error code: %" ErrorEbmPrintf, err);
    }
    const size_t cSamples = static_cast<size_t>(countSamples); // we trust our internal code that this is convertible
 
@@ -985,7 +1001,7 @@ SEXP CreateInteractionDetector_R(SEXP dataSetWrapped, SEXP bag, SEXP initScores)
    if(NILSXP != TYPEOF(bag)) {
       const size_t cSamplesVerify = static_cast<size_t>(CountInts(bag));
       if(cSamples != cSamplesVerify) {
-         error("CreateInteractionDetector_R cSamples != cSamplesVerify");
+         Rf_error("CreateInteractionDetector_R cSamples != cSamplesVerify");
       }
 
       aBag = reinterpret_cast<BagEbm *>(R_alloc(cSamples, static_cast<int>(sizeof(BagEbm))));
@@ -999,7 +1015,7 @@ SEXP CreateInteractionDetector_R(SEXP dataSetWrapped, SEXP bag, SEXP initScores)
       do {
          const int32_t replication = *pSampleReplicationR;
          if(IsConvertError<BagEbm>(replication)) {
-            error("CreateInteractionDetector_R IsConvertError<BagEbm>(replication)");
+            Rf_error("CreateInteractionDetector_R IsConvertError<BagEbm>(replication)");
          }
          if(0 != replication) {
             ++cExpectedInitScores;
@@ -1015,7 +1031,7 @@ SEXP CreateInteractionDetector_R(SEXP dataSetWrapped, SEXP bag, SEXP initScores)
       const IntEbm countInitScores = CountDoubles(initScores);
       size_t cInitScores = static_cast<size_t>(countInitScores);
       if(cInitScores != cExpectedInitScores) {
-         error("CreateInteractionDetector_R cInitScores != cExpectedInitScores");
+         Rf_error("CreateInteractionDetector_R cInitScores != cExpectedInitScores");
       }
       aInitScores = REAL(initScores);
    }
@@ -1025,11 +1041,14 @@ SEXP CreateInteractionDetector_R(SEXP dataSetWrapped, SEXP bag, SEXP initScores)
       pDataSet,
       aBag,
       aInitScores,
+      CreateInteractionFlags_Default,
+      AccelerationFlags_ALL,
+      "log_loss",
       nullptr,
       &interactionHandle
    );
    if(Error_None != err || nullptr == interactionHandle) {
-      error("CreateInteractionDetector returned error code: %" ErrorEbmPrintf, err);
+      Rf_error("CreateInteractionDetector returned error code: %" ErrorEbmPrintf, err);
    }
 
    SEXP interactionHandleWrapped = R_MakeExternalPtr(static_cast<void *>(interactionHandle), R_NilValue, R_NilValue); // makes an EXTPTRSXP
@@ -1046,39 +1065,46 @@ SEXP FreeInteractionDetector_R(SEXP interactionHandleWrapped) {
    return R_NilValue;
 }
 
-SEXP CalcInteractionStrength_R(SEXP interactionHandleWrapped, SEXP featureIndexes, SEXP minSamplesLeaf) {
+SEXP CalcInteractionStrength_R(SEXP interactionHandleWrapped, SEXP featureIndexes, SEXP maxCardinality, SEXP minHessian) {
    EBM_ASSERT(nullptr != interactionHandleWrapped); // shouldn't be possible
    EBM_ASSERT(nullptr != featureIndexes); // shouldn't be possible
-   EBM_ASSERT(nullptr != minSamplesLeaf);
+   EBM_ASSERT(nullptr != maxCardinality);
+   EBM_ASSERT(nullptr != minHessian);
 
    if(EXTPTRSXP != TYPEOF(interactionHandleWrapped)) {
-      error("CalcInteractionStrength_R EXTPTRSXP != TYPEOF(interactionHandleWrapped)");
+      Rf_error("CalcInteractionStrength_R EXTPTRSXP != TYPEOF(interactionHandleWrapped)");
    }
    const InteractionHandle interactionHandle = static_cast<InteractionHandle>(R_ExternalPtrAddr(interactionHandleWrapped));
    if(nullptr == interactionHandle) {
-      error("CalcInteractionStrength_R nullptr == interactionHandle");
+      Rf_error("CalcInteractionStrength_R nullptr == interactionHandle");
    }
 
    const IntEbm cDimensions = CountDoubles(featureIndexes);
    const IntEbm * const aFeatureIndexes = ConvertDoublesToIndexes(cDimensions, featureIndexes);
 
-   const IntEbm samplesLeafMin = ConvertIndexApprox(minSamplesLeaf);
+   const IntEbm cardinalityMax = ConvertIndexApprox(maxCardinality);
+   const double hessianMin = ConvertDouble(minHessian);
 
    double avgInteractionStrength;
    const ErrorEbm err = CalcInteractionStrength(
       interactionHandle, 
       cDimensions, 
       aFeatureIndexes, 
-      InteractionFlags_Default, 
-      samplesLeafMin, 
+      CalcInteractionFlags_Default, 
+      cardinalityMax, 
+      0,
+      hessianMin,
+      0,
+      0,
+      0,
       &avgInteractionStrength
    );
    if(Error_None != err) {
-      error("CalcInteractionStrength returned error code: %" ErrorEbmPrintf, err);
+      Rf_error("CalcInteractionStrength returned error code: %" ErrorEbmPrintf, err);
    }
 
-   SEXP ret = PROTECT(allocVector(REALSXP, R_xlen_t { 1 }));
-   REAL(ret)[0] = static_cast<double>(avgInteractionStrength);
+   SEXP ret = PROTECT(Rf_allocVector(REALSXP, R_xlen_t { 1 }));
+   REAL(ret)[0] = avgInteractionStrength;
    UNPROTECT(1);
    return ret;
 }
@@ -1104,7 +1130,7 @@ static const R_CallMethodDef g_exposedFunctions[] = {
    { "GetCurrentTermScores_R", (DL_FUNC)&GetCurrentTermScores_R, 2 },
    { "CreateInteractionDetector_R", (DL_FUNC)&CreateInteractionDetector_R, 3 },
    { "FreeInteractionDetector_R", (DL_FUNC)&FreeInteractionDetector_R, 1 },
-   { "CalcInteractionStrength_R", (DL_FUNC)&CalcInteractionStrength_R, 3 },
+   { "CalcInteractionStrength_R", (DL_FUNC)&CalcInteractionStrength_R, 4 },
    { NULL, NULL, 0 }
 };
 
